"""
Abstract Base class ParserABC for implementing transcription parsers
Underlies all parsers such as sbvParser, srtParser etc. 
"""
import abc
from textgrid_convert_ina import textgridtools as tgtools
from textgrid_convert_ina import preproctools as pptools
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class ParserABC(metaclass=abc.ABCMeta):
    """
    Abstract base class for Parsers to feed textgrid conversion
    Attributes:
        unique_id (str) : unique identifier for Parser
        transcription(str) : raw string of transcription
        transcription_dict(dict) : parsed form of transcription, of form  {$CHUNK_ID: {"start": int, "end": int, "text": str}}
    """
    unique_id = None
    transcription = None
    transcription_dict = None


    @abc.abstractmethod
    def parse_timestamp(self, timestamp):
        """
        Convert `timestamp` to int.

        Args:
            timestamp(str): string of timestamp, e.g. 0:0:23
        Returns:
            timestamp in milliseconds, e.g. 23000
        """

    @abc.abstractmethod
    def parse_transcription(self, transcription):
        """
        Convert raw transcription input string to transcription dictionary

        Args:
            transcription(str): raw string of transcription
        Returns:
            dict
        """

    def to_textgrid(self, input_dict=None, output_file=None, speaker_name="Speaker1", adapt_endstamps=0.001):
        """
        FIXME: add output_file
        Convert internal dict to Praat Textgrid format, as defined here: http://www.fon.hum.uva.nl/praat/manual/Intro_7__Annotation.html

        Args:
            speaker_name (str): name of speaker, default to "Speaker 1"
            adapt_endstamps(float): if given, will adapt end stamps to < start stamp of next chunk
        Returns:
            TextGrid compatible string
        """
        textgrid_dict = {}
        if input_dict is None:
            input_dict = self.transcription_dict
        # if not given or empty, default to self.transcription_dict`
        if len(input_dict) < 1:
            log.debug("No transcription dict found, running parse_transcription()")
            self.parse_transcription(self.transcription)
            input_dict = self.transcription_dict
        log.debug("Input dict has %s items" %len(input_dict))
        for chunk, values in input_dict.items():
            # FIXME: maybe this needs to be a parser implemented method
            start, end = values["start"], values["end"]
            textgrid_dict[chunk] = values
            textgrid_dict[chunk]["start"], textgrid_dict[chunk]["end"] = tgtools.ms_to_textgrid(start),  tgtools.ms_to_textgrid(end)
        # fix timestamp overlaps
        if adapt_endstamps:
            log.debug("Adapting end stamps with gap %f" %adapt_endstamps)
            textgrid_dict = pptools.adapt_timestamps(textgrid_dict, gap=adapt_endstamps)
        # FIXME this should be flexible in regards to long or short output format
        textgrid = tgtools.to_long_textgrid(tier_dict=textgrid_dict)
        return textgrid

    @classmethod
    def from_file(cls, input_path):
        """
        Read file from disk

        Args:
            input_path(str): path to transcription file to read
        Returns:
            initialized class
        """
        with open(input_path, "r", encoding="utf-8") as transcriptin:
            transcript = transcriptin.read()
        return cls(transcript, unique_id=input_path)

    def to_file(self):
        """
        Write file to disk
        """
        raise NotImplementedError("")





# read the sbv stuff in; assign timestamp to text
# {id : "chunk", "start", "end"
# expected out: 26 	 Carrie: 	 [78.28] 	 (pause 5.83) 	 [84.10]
import re
import logging
from textgrid_convert_ina.ParserABC import ParserABC
log = logging.getLogger(__name__)
log.addHandler(logging.StreamHandler())
#log.setLevel(logging.DEBUG)

class sbvParser(ParserABC):
    """
    Read and parse an sbv formatted file
    Inofficial specs here: GGL 

    Attributes:
        file_name(optional)
        sbv_text(str)
    """
    def __init__(self, transcription, unique_id=None):
        """
        Initializer

        Args:
            str_text(str)
        """
        self.transcription = transcription
        self.transcription_dict = {}
        self.parse_transcription(transcription)
        if unique_id is not None:
            self.unique_id=unique_id

    def parse_timestamp(self, timestamp):
        """
        Convert timestamps from sbv format 0:00:00.599 to ms
        """
        hrs, mins, rest = timestamp.split(":")
        hrs, mins = int(hrs), int(mins)
        secs, ms = [int(i) for i in rest.split(".")]
        fulltime = (hrs * 3600000) + (mins * 60000) + (secs * 1000) + ms
        assert isinstance(fulltime, int)
        return fulltime

    # FIXME: re compile
    def sbv_textparse(self, speaker_and_text, speaker="Speaker 1", speaker_regex=re.compile("[A-Z]+:")):
        """
        Args:
            speaker_and_text(str)
        Returns:
            tuple (SPEAKER(str), text(str))
        """
        raw_text = speaker_and_text.lstrip(">")
        new_speaker = speaker_regex.search(raw_text.lstrip())
        if new_speaker: 
            speaker = new_speaker.group()
        text = re.sub("^" + speaker, "", raw_text)
        text = text.strip()
        return speaker.rstrip(":"), text


    def parse_transcription(self, transcription, time_stamp_sep=","):
        """
        Pull the stuff from sbv into a dictionary of format {chunk_id: {
        "speaker": str, 
        "text": str, 
        "start": int, 
        "end": int}}

        Args:
            transcription(str) : 
            time_stamp_sep (str) : 
        Returns:
            dict as described above
        """
        chunk_id = 0
        for timestamps, speaker_and_text in self.sbv_generator(self.transcription.splitlines(), separator=""):
            start, end = timestamps.split(time_stamp_sep)
            previous_entry = self.transcription_dict.get(chunk_id, {"speaker_name": "Speaker 1"})
            speaker, text = self.sbv_textparse(speaker_and_text, speaker=previous_entry["speaker_name"])
            chunk_id += 1
            self.transcription_dict[chunk_id] = {
                    "speaker_name": speaker,
                    "start": self.parse_timestamp(start),  
                    "end": self.parse_timestamp(end),  
                    "text": text}
            log.debug("Added %s for chunk %s" %(self.transcription_dict[chunk_id], chunk_id))
        return self.transcription_dict.copy()

    def sbv_generator(self, filein, separator=""):
        """
        Args:
            filein(file read object or other iterable)
            separator(str): separator between records
        Returns:
            generator over chunk_id, timestamp, text
            FIXME: deque here
        """
        count = 0
        output = ()
        for line in filein:
            count +=1
            output = output + (line, )
            if count % 3 == 0:
                yield output[:-1]
                output = ()
        log.debug("sbv generator processed {} lines from {}".format(count, filein[:100]))


    def to_darla_textgrid(self, speaker_id=None, speaker_name=None, alias="sentence"):
        """
        Change TextGrid to the format DARLA understands: only "sentence" grids

        Args:
            speaker_id(int): NA for sbvs
            speaker_name(str): name of the speaker to extact
            alias: the name to use for texttier -- DARLA wants 'sentence'
        Returns:
            str to be fed into DARLA
        """
        if not self.transcription_dict:
            log.debug("Running parse_transcription for %s" %self.unique_id)
            self.parse_transcription()
        darla_dict = dict(self.transcription_dict)
        output_dict = dict(darla_dict)
        if speaker_name is not None:
            log.debug("Speaker name set to '%s'" %speaker_name)
            output_dict = {k:v for k,v in output_dict.items() if v["speaker_name"] == speaker_name}
            if not output_dict:
                existing_speakers = {v["speaker_name"] for k,v in output_dict.items()}
                raise ValueError("Speaker name '{}' not found in set of speakers '{}'".format(speaker_name, set(existing_speakers)))
        log.debug("Transcription dict going into DARLA has {} items".format(len(darla_dict)))
        log.debug("Setting tier name to '%s'" %alias)
        for key, values in darla_dict.items():
            output_dict[key] = values
            output_dict[key]["speaker_name"]  = alias
        textgrid = self.to_textgrid(output_dict, speaker_name="IO")
        return textgrid




# main
import os
import pathlib
import sys
from textgrid_convert_ina import iotools
from textgrid_convert_ina.ArgParser import  arg_parser
from textgrid_convert_ina.sbvParser import sbvParser
from textgrid_convert_ina.srtParser import srtParser
from textgrid_convert_ina.revParser import revParser
import logging

log = logging.getLogger(__name__)
log.addHandler(logging.StreamHandler())
log.setLevel(logging.DEBUG)

HERE = pathlib.Path(os.getcwd())

FILE_EXT_TO_FORMAT = {
    "json": "rev",
    "srt": "srt",
    "sbv": "sbv"
}


def get_py_version():
    """
    Get currently running Py version. Need to make pathlib adapts for Py < 3.6

    Args:

    Returns:
        tuple (major, minor, micro)
    """
    version_tuple = sys.version_info
    return version_tuple[:3]

def guess_source_format(input_path, extension_map=FILE_EXT_TO_FORMAT):
    """
    Based on file extension of `input_path`, guess the format of transcription file.

    Args:
        input_path(str): file name
        extension_map(dict): dictoinary {file_ext: format}, e.g. {"srt": "srt"}
    Returns:
        format string, None if not found in `extension_map`
    """ 
    _, ext = os.path.splitext(input_path)
    ext = ext.strip(".")
    source_format = extension_map.get(ext)
    if source_format is None:
        log.warning("Cannot guess source format for file '%s' (options are: %s)" %(input_path, extension_map.keys()))
    return source_format

def convert_to_txtgrid(input_file, source_format, speaker_name="Speaker 1"):
    """
    Convert from `source_format` in `input_file` to TextGrid.

    Args:
        input_file(str): path to input srt or sbv file to read from
        source_format(str): either sbv or srt
        speaker_name(str): optional speaker name
    Returns:
        TextGrid formatted string
    """
    with open(input_file, "r", encoding="utf-8") as sourcefile:
        sourcetext = sourcefile.read()
        log.debug("Read file '{}' with {} chars".format(input_file, len(sourcetext)))
    if source_format == "sbv":
        parser = sbvParser(sourcetext)
    if source_format == "srt":
        parser = srtParser(sourcetext)
    if source_format in ["json", "rev"]:
        parser = revParser(sourcetext)
    txtgrid = parser.to_textgrid(speaker_name=speaker_name)
    log.debug("Created TextGrid for file '{}', len {}".format(input_file, len(sourcetext)))
    return txtgrid

def convert_to_darla(input_file, source_format, speaker_name="Speaker 1"):
    """
    Convert from `source_format` in `input_file` to DARLA-compatible TextGrid

    Args:
        input_file(str): path to input srt or sbv file to read from
        source_format(str): either sbv or srt
        speaker_name(str): optional speaker name
    Returns:
        TextGrid formatted string
    """
    with open(input_file, "r", encoding="utf-8") as sourcefile:
        sourcetext = sourcefile.read()
        log.debug("Read file '{}' with {} chars".format(input_file, len(sourcetext)))
    if source_format == "sbv":
        parser = sbvParser(sourcetext)
    if source_format == "srt":
        parser = srtParser(sourcetext)
    if source_format in ["json", "rev"]:
        parser = revParser(sourcetext)
    # this will default to first speaker for rev
    txtgrid = parser.to_darla_textgrid(alias="sentence")
    log.debug("Created DARLA TextGrid for file '{}', len {}".format(input_file, len(sourcetext)))
    return txtgrid

def folder_source_format(input_folder, file_types=[".srt", ".sbv", ".json", ".rev"]):
    """
    Check whether files in `input_foldelibr` have sbv, srt endings

    Args:
        input_folder(str)
        file_types (iterable of str): file endings to consider
    Returns:
        str srt or sbv
    Raises:
        ValueError if mix of extensions
    """
    # FIXME: this should use guess_format
    input_folder = str(input_folder)
    infiles = [i for i in os.listdir(input_folder)]
    infiles = [i for i in infiles if os.path.splitext(i.lower())[-1] in file_types]
    types = [os.path.splitext(i.lower())[-1]for i in infiles]
    if len(infiles) < 1:
        raise ValueError("No relevant sbv or srt files found in folder '{}', contains '{}'".format(
            input_folder, os.listdir(input_folder)[:100]))
    if len(set(types)) > 1:
        raise ValueError("Both sbv and srt files found in folder '{}', contains '{}'".format(
            input_folder, infiles[:100]))
    assert types[0] in file_types
    #FIXME do we really need the check above
    return types[0].lstrip(".")

def main(source_format, to,  input_path, output_path=HERE, suffix="_TEXTGRID.txt", strict=True):
    """
    Convert files(s) from `input_path` from  `to` format to TextGrid. Optionally, write to `output_path`
    Example: convert from=sbv to=TextGrid and write to output_path="home/patrick/output"

    Args:
        source_format(str) : transcription format, currently accepts sbv, srt, and rev
        to (str) : file ending, only accepts TextGrid atm
        input_path(str)
        output_path(str)
        suffix(str): string to append to file name for writing out TextGrid
        strict(Bool): if True, will not overwrite files
    """
    log.debug('strict is %s' %strict)
    log.debug('HERE is %s' %HERE)
    major, minor, micro = get_py_version()
    if major != 3:
        version_str = ".".join([str(major), str(minor), str(micro)])
        raise NotImplementedError("Textgrid-convert only works with Python 3, not '{}'".format(version_str))
    if minor < 5:
        input_path = str(input_path)
        output_path = str(output_path)
    else:
        input_path = pathlib.Path(input_path)
    log.debug('input path %s, output path %s' %(input_path, output_path))
    if source_format:
        source_format = source_format.lower().strip(" .")
        if source_format not in FILE_EXT_TO_FORMAT.values():
            raise ValueError("Works with '{}'  given '{}'".format(FILE_EXT_TO_FORMAT.values(), source_format))
    if not any([source_format, input_path]):
        raise ValueError("Either input_path or source format needs to be specified, currently are {} and {}".format(
            source_format, input_path))
    # processing FIXME: outsource this
    if os.path.isdir(input_path):
        log.debug("Processing folder '{}'".format(str(input_path)))
        if not os.path.isdir(output_path):
            raise IOError("Cannot write multiple output files to path '{}' (specify output folder with `-o`)".format(
                output_path))
        if not source_format:
            source_format = folder_source_format(input_path)
            #FIXME below
        file_ext = {v:k for k,v in FILE_EXT_TO_FORMAT.items()}.get(source_format)
        if file_ext is None:
            raise KeyError("Could not find file type associated with '{}', options are: '{}'".format(
                source_format, FILE_EXT_TO_FORMAT.values()))
        infiles = [i for i in os.listdir(input_path) if os.path.splitext(i.lower())[-1] == "." + file_ext]
        infiles = [os.path.join(input_path, i) for i in infiles]
        if len(infiles) < 1:
            raise ValueError("No relevant files found in folder '{}', contains '{}'".format(
                input_path, os.listdir(input_path)[:100]))
        log.debug("Processing {} {} files from folder '{}'".format(len(infiles), source_format, str(input_path)))
        for fil in infiles:
            log.debug("Working on file '{}'".format(fil))
            if to and to.lower() in ["darla", "darlatextgrid"]:
                log.debug("Convert file '{}' to DARLA TextGrid".format(fil))
                result = convert_to_darla(fil, source_format)
            else:
                result = convert_to_txtgrid(fil, source_format)
            if output_path:
                #filename = output_path / (fil.name + suffix)
                fil_name = os.path.split(fil)[-1]
                filename = os.path.join(output_path, fil_name + suffix)
                log.debug("Writing to %s" %filename)
                iotools.filewriter(filename, outstring=result, strict=strict)
    else:
        log.debug("Working on file '%s'" %input_path)
        if not source_format:
            source_format = guess_source_format(input_path)
            if not source_format:
                raise IOError("Don't know what source format file '%s' is. Specify with option `-f`", input_path)
        if to and to.lower() in ["darla", "darlatextgrid"]:
            result = convert_to_darla(input_path, source_format)
        else:
            result = convert_to_txtgrid(input_path, source_format)
        if not output_path:
            output_path = HERE 
        fil_name = os.path.split(input_path)[-1]
        if os.path.isdir(output_path):
            print("given a folder")
            filename = os.path.join(output_path, fil_name + suffix)
        else:
            filename = output_path
        log.debug("Writing to %s", filename)
        iotools.filewriter(filename, outstring=result, strict=strict)
        log.debug("Written to %s", filename)

if __name__ == "__main__":
    current_args = arg_parser.parse_args()
    log.debug("Current args %s " %current_args)
    main(**vars(current_args))



import logging
log = logging.getLogger(__name__)
log.addHandler(logging.StreamHandler())

from textgrid_convert_ina.ttextgrid_convert import main
from textgrid_convert_ina.ArgParser import  arg_parser

current_args = arg_parser.parse_args()
log.debug("Current args %s " %current_args)
main(**vars(current_args))



"""
Data preprocessing tools. 
"""

from collections import OrderedDict
import logging
import copy
log = logging.getLogger(__name__)

def adapt_timestamps(input_dict, gap=0.1):
    """
    Adapt time end stamps to not overlap with following start time stamp.

    Args:
        input_dict(dict) : dictionary with timestamps, e.g. self.transcription_dict in a Parser
        gap(float, optional): gap to introduce between end and start index after adapt. Defaults to 0.1
    Returns:
        dict
    """
    chunks = copy.deepcopy(input_dict)
    # FIX ME this should probably be an Ordereddict to begin with
    sorted_chunks = OrderedDict(sorted(chunks.items(), key=lambda x:
                                       x[1].get("start")))
    keys = list(sorted_chunks.keys())
    for ind, key in enumerate(keys):
        if ind + 1== len(keys):
            break
        start, end = sorted_chunks[key].get("start"), sorted_chunks[key].get("end")
        next_key = keys[ind+1]
        next_start = sorted_chunks[next_key].get("start")
        if end >= next_start:
            new_end = max(next_start - gap, start)
            sorted_chunks[key]["end"] = new_end
            log.debug("Updated key '%s' end from %f to %f" %(key, end, new_end))
    return sorted_chunks



# read the srt stuff in; assign timestamp to text
# {id : "chunk", "start", "end"
# expected out: 26 	 Carrie: 	 [78.28] 	 (pause 5.83) 	 [84.10]
import logging
import copy
from textgrid_convert_ina.ParserABC import ParserABC

log = logging.getLogger(__name__)
from textgrid_convert_ina.textgridtools import merge_text_with_newlines

class srtParser(ParserABC):
    """
    Read and parse an srt formatted file
    Inofficial specs here: http://forum.doom9.org/showthread.php?p=470941#post470941

    """

    def __init__(self, transcription, preprocessors=[merge_text_with_newlines], unique_id=None):
        """
        Initializer

        Args:
            str_text(str)
        """
        self.preprocessors = preprocessors
        if self.preprocessors:
            for prep in self.preprocessors:
                self.transcription = prep(transcription)
        else:
            self.transcription = transcription
        self.transcription_dict = {}
        self.parse_transcription(self.transcription)
        if unique_id is not None:
            self.unique_id=unique_id


    def parse_timestamp(self, timestamp):
        """
        Convert from srt style timestamp 00:59:58,89 to ms

        Args:
            timestamp(str)
        Return:
            int
        """
        time, ms = timestamp.split(",")
        hrs, mins, secs = [int(i) for i in time.split(":")]
        fulltime = (hrs * 3600000) + (mins * 60000) + (secs * 1000) + int(ms)
        return fulltime


    def parse_transcription(self, srt_text=None, speaker_name="Speaker 1", time_stamp_sep=" --> "):
        """
        Pull the stuff from srt into a dictionary of format {chunk_id: {"text": "", "start": int, "end": int}}

        Args:
            srt_text(str): 
            speaker_name(str): 
            time_stamp_sep(str): placeholder between start and end time stamp
        Returns:
            dict as described above
        """
        if not srt_text:
            srt_text = self.transcription
        for chunk_id, timestamps, text in self.srt_generator(srt_text.splitlines()):
            start, end = timestamps.split(time_stamp_sep)
            self.transcription_dict[chunk_id] = {
                    "speaker_name": speaker_name,
                    "start": self.parse_timestamp(start),
                    "end": self.parse_timestamp(end),
                    "text": text}
        return copy.deepcopy(self.transcription_dict)

    def srt_generator(self, filein):
        """
        Args:
            filein(file read object or other iterable)
        Returns:
            generator over chunk_id, timestamp, text
        """
        count = 0
        output = ()
        for line in filein:
            count +=1
            output = output + (line, )
            if count % 4 == 0:
                yield output[:-1]
                output = ()
        log.debug("srt generator processed {} lines from {}".format(count, filein))

    def to_darla_textgrid(self, speaker_id=None, speaker_name=None, alias="sentence"):
        """
        Change TextGrid to the format DARLA understands: only "sentence" grids

        Args:
            speaker_id(int): NA for sbvs
            speaker_name(str): name of the speaker to extact
            alias: the name to use for texttier -- DARLA wants 'sentence'
        Returns:
            str to be fed into DARLA
        """
        if not self.transcription_dict:
            log.debug("Running parse_transcription for %s" %self.unique_id)
            self.parse_transcription()
        darla_dict = dict(self.transcription_dict)
        output_dict = dict(darla_dict)
        if speaker_name is not None:
            log.debug("Speaker name set to '%s'" %speaker_name)
            output_dict = {k:v for k,v in output_dict.items() if v["speaker_name"] == speaker_name}
            if not output_dict:
                existing_speakers = {v["speaker_name"] for k,v in output_dict.items()}
                raise ValueError("Speaker name '{}' not found in set of speakers '{}'".format(speaker_name, set(existing_speakers)))
        log.debug("Transcription dict going into DARLA has {} items".format(len(darla_dict)))
        log.debug("Setting tier name to '%s'" %alias)
        for key, values in darla_dict.items():
            output_dict[key] = values
            output_dict[key]["speaker_name"] = alias
        textgrid = self.to_textgrid(output_dict)
        return textgrid



"""
Functionality to parse CLI arguments.
"""
import pathlib
import argparse
import os
import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.StreamHandler())


HERE = pathlib.Path(os.getcwd())
# accepted format needs to be "REV" no JSON

ACCEPTED_INPUT_FORMATS = ["rev", "sbv", "srt", "json"]
ACCEPTED_INPUT_FORMATS = ACCEPTED_INPUT_FORMATS + [i.upper() for i in ACCEPTED_INPUT_FORMATS]
ACCEPTED_INPUT_FORMATS = ACCEPTED_INPUT_FORMATS + [i + "." for i in ACCEPTED_INPUT_FORMATS]
log.debug("Accepted input formats are %s" %ACCEPTED_INPUT_FORMATS)

"""
Flags:
    -f, --from: sbv or srt # optional: can also be inferred from file name
    -t, --to: txtgrid 
    -i, --input_path:  also folder mode?
    -o, --output_path #take files or folders
"""

arg_parser = argparse.ArgumentParser(description="convert srt and sbv files to Praat textgrid", allow_abbrev=True)

"""
Set up read write and convert arguments
"""
arg_parser.add_argument("--input_path", "-i", type=str, help="Path to input file or folder", dest="input_path", required=True )
arg_parser.add_argument("--output_path", "-o",  help="Path to write output file(s) to", dest="output_path", default=HERE, type=pathlib.PurePath)
arg_parser.add_argument("--source_format", "--from", "-f", type=str, help="Input format to convert to TextGrid, e.g. sbv or srt",
        dest="source_format", choices=ACCEPTED_INPUT_FORMATS)
arg_parser.add_argument("--to", "-t", type=str, help="Output format to convert to, e.g. TextGrid or Darla", dest="to" )
arg_parser.add_argument("--strict",  action="store_false", help="If set, will overwrite existing files", dest="strict")
arg_parser.add_argument("--overwrite",  action="store_false", help="If set, will overwrite existing files", dest="strict")



"""
Collect TextGrid related functionality here
"""
import re
import logging
log = logging.getLogger(__name__)

# this is what we have in sbvParse

# Pull the stuff from sbv into a dictionary of format 
# {chunk_id: {"text": "", "start": int, "end": int}}

testdict = {0: {"speaker": "Mary", "text": "one", "start": 0, "end": 0.5896534423132239},
            1: {"speaker": "Mary", "text": "",  "end": 1.4123177579131596,"start": 0.5896534423132239},
            2: {"speaker": "Mary", "text": "two",  "end": 2.343227378197297, "start": 1.4123177579131596},
            3: {"speaker": "Mary", "text": "three",  "end": 3.1225935719235522, "start": 2.34322737819729},
            4: {"speaker": "Mary", "start": 3.1225935719235522, "end": 12.804852607709751, "text": "rest of the text * @ " 
               },
            5: {"speaker": "John", "start": 0, "end": 12.804852607709751, "text": "" }}


def ms_to_textgrid(milliseconds, strict=True):
    """
    Convert milliseconds to textgrid appropriate format,e.g. 12.88

    Args:
        milliseconds (int)
        strict(Bool): if True, will error out if no int given
    """
    if not isinstance(milliseconds, int):
        raise ValueError("Value {} for milliseconds is not an integer, is {}".format(milliseconds, type(milliseconds)))
    return milliseconds / 1000


def to_textgrid_time(timestamp, split_char="."):
    """
    FIXME: deprecate
    Output needs to be in mili seconds, round to 2

    Args:
        timestamp(str)
    Returns 
    """
    if not isinstance(timestamp, str):
        log.debug("Converting timestamp {} to string".format(timestamp))
        timestamp = str(timestamp)
    time, ms = timestamp.split(split_char)
    hours, mins, secs = [float(i) for i in time.split(":")]
    fulltime = (hours * 3600000) + (mins * 60000) + (secs * 1000) + float(ms)
    fulltime = fulltime / 1000
    return fulltime

TEXTGRID_HEADER = """
File type = "ooTextFile"
Object class = "TextGrid"
"""

def to_short_textgrid(tier_dict):
    """
    """
    return 1


def collect_chunk_values(input_dict, key, strict=True):
    """
    Collect all values associated with chunks in input_dict.

    Args:
        input_dict(dict): {chunk_id: {key: value}}
        key(str)
        strict(Bool): if True, will error out if `key` not present
    Returns:
        list of results
    """
    results = [] 
    for input_key, values in input_dict.items():
        result = values.get(key, "NOT FOUND")
        if strict and result == "NOT FOUND":
            raise KeyError("Key '{}' not found, has {}".format(key, values.keys()))
        results.append(result)
    return results




def to_long_textgrid(tier_dict, tier_key="speaker_name", tier_class="IntervalTier"):
    """
    Create long form TextGrid, cf specs here:http://www.fon.hum.uva.nl/praat/manual/TextGrid_file_formats.html.

    Args:
        tier_dict(dict): 
        tier_key(str) :  
        tier_class(str) :  tier class to use for TextGrid
    Returns:
        str of TextGrid 
    """
    tier_names = set(collect_chunk_values(tier_dict, tier_key))
    tier_times = collect_chunk_values(tier_dict, "start") + collect_chunk_values(tier_dict, "end")
    tier_data = list(tier_dict.values())
    # make the file header 
    file_header = "File type = \"ooTextFile\"\nObject class = \"TextGrid\"\n"
    file_min = "xmin = {}".format(min(tier_times))
    file_max = "xmax = {}".format(max(tier_times))
    file_tiers_flag = "tiers? <exists>"
    file_size = "size = {}".format(len(tier_names))
    file_item = "item []:"
    file_str = "\n".join([file_header, file_min, file_max, file_tiers_flag, file_size, file_item])
    tiers = []
    # note that the content is == as the short version, only we do the string formatting
    for ind, name in enumerate(sorted(tier_names, reverse=True)):
        name_data = [i for i in tier_data if i[tier_key] == name] 
        name_data_sorted = sorted(name_data, key = lambda x: x["start"])
        log.debug("{} items found for tier name '{}'".format(len(name_data), name))
        # make the header over intervals
        tier_header = "\titem [{}]:".format(ind + 1)
        tier_metadata =  "\t\tclass = \"{}\"".format(tier_class)
        tier_name = "\t\tname = \"{}\"".format(name)
        tier_min = "\t\txmin = {}".format(min([val["start"] for val in name_data_sorted]))
        tier_max = "\t\txmax = {}".format(max([val["end"] for val in name_data_sorted]))
        n_tier_intervals = "\t\tintervals: size = {}\n".format(len(name_data_sorted))
        tier_intro = "\n".join([tier_header, tier_metadata, tier_name, tier_min, tier_max, n_tier_intervals])
        # make the intervals
        tier_intervals = []
        for ind, interval in enumerate(name_data_sorted): 
            #FIXME the tabs might go into {}
            interval_header = "\t\t\tintervals [{}]:".format(ind + 1)
            interval_min = "\t\t\t\txmin = {}".format(interval["start"])
            interval_max = "\t\t\t\txmax = {}".format(interval["end"])
            interval_text = "\t\t\t\ttext = \"{}\"".format(interval["text"])
            interval_str = "\n".join([interval_header, interval_min, interval_max, interval_text])
            tier_intervals.append(interval_str)
        tiers = tiers + [tier_intro + "\n".join(tier_intervals)]
    return "\n".join([file_str, "\n".join(tiers)]) 

def merge_text_with_newlines(input_string):
    """
    Remove all empty lines from `input_string` as these can break the parser
    Args:
        input_string (str):
    Returns:
        input_string with empty lines dropped
    """
    separator = "\n"
    text_line_pattern = re.compile(r".*[A-Za-z].*")
    split_string = input_string.split(separator)
    output_split_string = []
    aggregated_strings = []
    # since the first line will always be an index, this will not wrap around
    for ind, line in enumerate(split_string):
        if not text_line_pattern.match(line):
            if aggregated_strings:
                output_split_string.append(" ".join(aggregated_strings))
                aggregated_strings = []
            output_split_string.append(line)
        else:
            aggregated_strings.append(line)
    if aggregated_strings:
        output_split_string.append(" ".join(aggregated_strings))
    return separator.join(output_split_string)


"""
Collect read and write functions.
"""
import os
import pytest
import logging
log = logging.getLogger(__name__)

def filewriter(filename, outstring, strict=True):
    """
    Write `outstring` to `filename`. 

    Args:
        filename(str) : path to file
        outstring(str) : string to write to file
        strict(Bool): if True, will not overwrite any existing files. Defaults to True
    """
    filename = str(filename)
    log.debug("Writing to %s", filename)
    if  os.path.isfile(filename):
        if strict:
            raise IOError(
                "File '{}' already exists, will not overwrite".format(str(filename)))
        else:
            log.warning(
                "File '%s' already exists, overwriting", str(filename))
            os.remove(filename)
    with open(filename, "w", encoding="utf-8") as gridout:
        gridout.write(outstring)
    log.debug("Textgrid written to '%s'", str(filename))



from textgrid_convert_ina.ParserABC import ParserABC
import copy
import json
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
log.addHandler(logging.StreamHandler())


def parse_revstamp(timestamp):
    """
    Convert timestamp from rev format (00:00:20,000) to ms

    Args:
        timestamp(str):
    Returns:
        int
    """
    hrs, mins, rest = timestamp.split(":")
    hrs, mins = int(hrs), int(mins) 
    secs, ms = [int(i) for i in rest.split(",")]
    fulltime = (hrs * 3600000) + (mins * 60000) + (secs * 1000) + ms
    fulltime = int(fulltime)
    assert isinstance(fulltime, int)
    return fulltime


class revParser(ParserABC):
    """
    Class to parse revfiles as defined here: https://www.rev.com/api/attachmentsgetcontent
    Attributes:
        speakers
    """
    speakers = () # pull info rom Rev JSON here


    def __init__(self, transcription, unique_id=None):
        """
        Initializer

        Args:
            transcription(str); transcription text
            unique_id(str): optional
        """
        self.transcription = transcription
        self.transcription_dict = {}
        self.parse_transcription(transcription)
        if unique_id is not None:
            self.unique_id=unique_id


    def parse_timestamp(self, timestamp):
        """
        Convert from rev timestamps to ms
        Args:
            timestamp(str): timestamp
        Returns:
            int of milliseconds
        """
        return parse_revstamp(timestamp)

    def parse_rev_monologues(self, monologues, chunk_id=1):
        """
        Parse clunky monologues into dict

        Args:
            monologues(iterable): list of dicts, as present in rev outout
            chunk_id (int): starting integer for chunk IDs, defaults to 1
        Returns:
            dictionary to be used for self.transcription_dict
        """
        chunk_id = 1
        mono_dict = {}
        log.debug("Found %s monologues", len(monologues))
        for mono in monologues:
            speaker_name, speaker_id = mono.get("speaker_name"), mono.get("speaker_id")
            utterances = [i for i in mono["elements"] if i["type"] == "text" and all((i.get("timestamp"), i.get("end_timestamp")))]
            log.debug("Found {} utterances".format(len(utterances)))
            for utter in utterances:
                mono_dict[chunk_id] = {}
                mono_dict[chunk_id]["speaker_name"], mono_dict[chunk_id]["speaker"] = speaker_name, speaker_id
                mono_dict[chunk_id]["text"] = utter["value"]
                mono_dict[chunk_id]["start"] = self.parse_timestamp(utter["timestamp"])
                mono_dict[chunk_id]["end"] = self.parse_timestamp(utter["end_timestamp"])
                chunk_id += 1
        return mono_dict

    def parse_transcription(self, speaker=None):
        """
        Read rev transcription into dict. Specs are here: https://www.rev.com/api/attachmentsgetcontent
        Args:
            speaker(str, optional): 
        """
        try:
            input_dict = json.loads(self.transcription)
        except json.decoder.JSONDecodeError as err:
            log.critical("Input transcription to revParser '%s' cannot be parsed as JSON" %self.unique_id) 
            raise err
        self.speakers = [(i["id"], i["name"]) for i in input_dict["speakers"]]
        monologues = input_dict["monologues"]
        monologue_dict = self.parse_rev_monologues(monologues)
        self.transcription_dict=monologue_dict
        return monologue_dict

    def to_darla_textgrid(self, speaker_id=None, alias="sentence"):
        """
        Change TextGrid to the format DARLA understands: only "sentence" grids

        Args:
            speaker_id (int):  ID of the speaker to keep, will default to first found
            alias (str): name of output tier, defaults to 'sentence' as required by DARLA
        Returns:
            str to be fed into DARLA
        """
        if not self.transcription_dict:
            log.debug("Running parse_transcription for %s" %self.unique_id)
            self.parse_transcription()
        if speaker_id is not None:
            if speaker_id not in [speaker_id for speaker_id, speaker_name in self.speakers]:
                raise ValueError("Speaker ID '{}' not found in set of speakers {}".format(speaker_id, self.speakers))
            _ , speaker_name = [(i, n) for i, n in self.speakers if i == speaker_id][0]
        else:
            speaker_id, speaker_name = self.speakers[0]
        log.debug("Speaker name set to '%s'" %speaker_name)
        darla_dict = copy.deepcopy(self.transcription_dict)
        output_dict = {k:v for k,v in darla_dict.items() if v["speaker_name"] == speaker_name}
        log.debug("Transcription dict going into DARLA has {} items".format(len(darla_dict)))
        log.debug("Setting tier name to '%s'" %alias)
        for key, values in darla_dict.items():
            output_dict[key] = values
            output_dict[key]["speaker_name"]  = alias
        textgrid = self.to_textgrid(output_dict, speaker_name="IO")
        return textgrid

